@@ -64,7 +64,7 @@
   :type 'string
   :group 'psgml-html)
 
-(defcustom html-helper-htmldtd-version "<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML//EN\">\n"
+(defcustom html-helper-htmldtd-version "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n"
   "*Version of HTML DTD you're using."
   :type 'string
   :group 'psgml-html)
@@ -131,13 +131,27 @@ of the function `html-helper-insert-timestamp' if
 ;; ones we know of.
 
 (defcustom html-helper-types-to-install
-  '(anchor header logical phys list textel entity image head form)
+  '(anchor header logical phys list textel entity image head form table
+	   special)
   "*List of tag types to install when html-helper-mode is first loaded.
 If you want to not install some type of tag, override this variable.
 Order is significant: menus go in this order."
   :type '(repeat symbol)
   :group 'psgml-html)
 
+(defcustom html-helper-use-expert-menu nil
+  "*If not nil, then use the full HTML menu."
+  :type 'boolean
+  :group 'psgml-html)
+
+(defcustom html-helper-user-menu nil
+  "*Extra items to put in the HTML expert menu.
+The value of this symbol is appended to the beginning of the expert
+menu that is handed off to easymenu for definition. It should be a
+list of vectors or lists which themselves are vectors (for submenus)."
+  :type 'sexp
+  :group 'psgml-html)
+
 ;;}}} end of user variables
 ;;{{{ type based keymap and menu variable and function setup
 
@@ -210,7 +224,7 @@ More specifically:
 
   ;; font-lock setup for various emacsen: XEmacs, Emacs 19.29+, Emacs <19.29.
   ;; By Ulrik Dickow <dickow@nbi.dk>.  (Last update: 05-Sep-1995).
-  (cond	((string-match "XEmacs\\|Lucid" (emacs-version)) ; XEmacs/Lucid
+  (cond	(running-xemacs ; XEmacs/Lucid
 	 (put major-mode 'font-lock-keywords-case-fold-search t))
 	;; XEmacs (19.13, at least) guesses the rest correctly.
 	;; If any older XEmacsen don't, then tell me.
@@ -243,13 +257,46 @@ More specifically:
   (modify-syntax-entry ?\\ ".   " html-mode-syntax-table)
   (modify-syntax-entry ?'  "w   " html-mode-syntax-table)
 
+  (tempo-use-tag-list 'html-helper-tempo-tags html-helper-completion-finder)
+  (setq imenu-create-index-function 'html-helper-imenu-index)
+  (setq imenu-sort-function nil) ; sorting the menu defeats the purpose
+
   ; sigh ...  need to call this now to get things working.
   (sgml-build-custom-menus)
   ;; (add-submenu nil sgml-html-menu "SGML")
+  (setq sgml-menu-name "HTML")
   (easy-menu-add sgml-html-menu)
+  (html-helper-rebuild-menu)
   (unless (featurep 'infodock)
     (delete-menu-item '("SGML"))))
 
+(defvar html-helper-imenu-regexp
+  "\\s-*<h\\([1-9]\\)[^\n<>]*>\\(<[^\n<>]*>\\)*\\s-*\\([^\n<>]*\\)"
+  "*A regular expression matching a head line to be added to the menu.
+The first `match-string' should be a number from 1-9.
+The second `match-string' matches extra tags and is ignored.
+The third `match-string' will be the used in the menu.")
+
+;; Make an index for imenu
+(defun html-helper-imenu-index ()
+  "Return an table of contents for an html buffer for use with Imenu."
+  (let ((space ?\ ) ; a char
+	(toc-index '())
+	toc-str)
+    (save-excursion
+      (goto-char (point-min))
+      (while (re-search-forward html-helper-imenu-regexp nil t)
+	(setq toc-str
+	      (concat
+	       (make-string
+		(* 2 (- (string-to-number (match-string 1)) 1))
+		space)
+	       (match-string 3)))
+	(beginning-of-line)
+	(setq toc-index (cons (cons toc-str (point)) toc-index))
+	(end-of-line)))
+    (nreverse toc-index)))
+
 (defun html-helper-add-type-to-alist (type)
   "Add a type specification to the alist.
 The spec goes (type . (keymap-symbol keyprefix menu-symbol menu-string)).
@@ -260,14 +307,16 @@ See code for an example."
 (mapcar 'html-helper-add-type-to-alist
   '((entity  . (nil nil html-helper-entity-menu "Insert Character Entities"))
     (textel  . (nil nil html-helper-textel-menu "Insert Text Elements"))
-    (head    . (html-helper-head-map "\C-zb" html-helper-head-menu "Insert Structural Elements"))
+    (head    . (html-helper-head-map "\C-zw" html-helper-head-menu "Insert Structural Elements"))
     (header  . (html-helper-base-map "\C-z" html-helper-header-menu "Insert Headers"))
     (anchor  . (html-helper-base-map "\C-z" html-helper-anchor-menu "Insert Hyperlinks"))
     (logical . (html-helper-base-map "\C-z" html-helper-logical-menu "Insert Logical Styles"))
     (phys    . (html-helper-base-map "\C-z" html-helper-phys-menu "Insert Physical Styles"))
     (list    . (html-helper-list-map "\C-zl" html-helper-list-menu "Insert List Elements"))
     (form    . (html-helper-form-map "\C-zf" html-helper-form-menu "Insert Form Elements"))
-    (image   . (html-helper-image-map "\C-zm" html-helper-image-menu "Insert Inlined Images"))))
+    (table   . (html-helper-table-map "\C-zt" html-helper-table-menu "Insert Table Elements"))
+    (image   . (html-helper-image-map "\C-zm" html-helper-image-menu "Insert Inlined Images"))
+    (special . (html-helper-base-map "\C-z" html-helper-special-menu "Insert Specials"))))
 
 ;; Once html-helper-mode is aware of a type, it can then install the
 ;; type: arrange for keybindings, menus, etc.
@@ -310,13 +359,13 @@ with html-helper-add-type-to-alist."
   "Given a string, downcase it and replace spaces with -.
 We use this to turn menu entries into good symbols for functions.
 It's not entirely successful, but fortunately emacs lisp is forgiving."
-  (let* ((s (copy-sequence input-string))
+  (let* ((s (downcase input-string))
 	 (l (1- (length s))))
-    (while (> l 0)
+    (while (>= l 0)
       (if (char-equal (aref s l) ?\ )
 	  (aset s l ?\-))
       (setq l (1- l)))
-    (concat "html-" (downcase s))))
+    (concat "html-" s)))
 
 
 (defun html-helper-add-tag (l)
@@ -352,9 +401,8 @@ appropriate keymap if a key is requested. Format:
 
 ;;{{{ most of the HTML tags
 
-;; These tags are an attempt to be HTML/2.0 compliant, with the exception
-;; of container <p>, <li>, <dd>, <dt> (we adopt 3.0 behaviour).
-;; For reference see <URL:http://www.w3.org/hypertext/WWW/MarkUp/MarkUp.html>
+;; These tags are an attempt to be HTML 3.2 compliant
+;; For reference see <URL:http://www.w3.org/TR/REC-html32.html>
 
 ;; order here is significant: within a tag type, menus and mode help
 ;; go in the reverse order of what you see here. Sorry about that, it's
@@ -364,92 +412,113 @@ appropriate keymap if a key is requested. Format:
  'html-helper-add-tag
  '(
    ;;entities
-   (entity  "\C-c#"   "&#"              "Ascii Code"      ("&#" (r "Ascii: ") ";"))
-   (entity  "\C-c\""  "&quot;"          "Quotation mark"  ("&quot;"))
-   (entity  "\C-c$"   "&reg;"           "Registered"      ("&reg;"))
-   (entity  "\C-c@"   "&copy;"          "Copyright"       ("&copy;"))
-   (entity  "\C-c-"   "&shy;"           "Soft Hyphen"     ("&shy;"))
-   (entity  "\C-c "   "&nbsp;"		"Nonbreaking Space"  ("&nbsp;"))
-   (entity  "\C-c&"   "&amp;"		"Ampersand"	  ("&amp;"))
-   (entity  "\C-c>"   "&gt;"	  	"Greater Than"       ("&gt;"))
-   (entity  "\C-c<"   "&lt;"		"Less Than"	  ("&lt;"))
+   (entity  "\C-c#"   "&#"              "Ascii Code"     ("&#" (r "Ascii: ") ";"))
+   (entity  "\C-c\""  "&quot;"          "Quotation mark" ("&quot;"))
+   (entity  "\C-c$"   "&reg;"           "Registered"     ("&reg;"))
+   (entity  "\C-c@"   "&copy;"          "Copyright"      ("&copy;"))
+   (entity  "\C-c-"   "&shy;"           "Soft Hyphen"    ("&shy;"))
+   (entity  "\C-c "   "&nbsp;"		"Nonbreaking Space" ("&nbsp;"))
+   (entity  "\C-c&"   "&amp;"		"Ampersand"	 ("&amp;"))
+   (entity  "\C-c>"   "&gt;"	  	"Greater Than"   ("&gt;"))
+   (entity  "\C-c<"   "&lt;"		"Less Than"	 ("&lt;"))
    
    ;; logical styles
-   (logical "q"       "<blockquote>"	"Blockquote"     	  ("<blockquote>" (r "Quote: ") "</blockquote>"))
-   (logical "c"       "<code>"		"Code"           	  ("<code>" (r "Code: ") "</code>"))
-   (logical "x"       "<samp>"		"Sample"         	  ("<samp>" (r "Sample code") "</samp>"))
-   (logical "r"       "<cite>"		"Citation"       	  ("<cite>" (r "Citation: ") "</cite>"))
-   (logical "k"       "<kbd>"		"Keyboard Input"       	  ("<kbd>" (r "Keyboard: ") "</kbd>"))
-   (logical "v"       "<var>"		"Variable"       	  ("<var>" (r "Variable: ") "</var>"))
-   (logical "d"       "<dfn>"		"Definition"     	  ("<dfn>" (r "Definition: ") "</dfn>"))
-   (logical "a"	      "<address>"	"Address"		  ("<address>" r "</address>"))
-   (logical "e"       "<em>"		"Emphasized"     	  ("<em>" (r "Text: ") "</em>"))
-   (logical "s"       "<strong>"	"Strong"         	  ("<strong>" (r "Text: ") "</strong>"))
-   (logical "p"       "<pre>"		"Preformatted"   	  ("<pre>" (r "Text: ") "</pre>"))
+   (logical "v"       "<div"            "Text Division"  ("<div align=\"" (r "Alignment: ") "\">" (r "Text: ") "</div>"))
+   (logical "n"       "<center>"        "Center"         ("<center>" (r "Text: ") "</center>"))
+   (logical "q"       "<blockquote>"	"Blockquote"     ("<blockquote>" (r "Quote: ") "</blockquote>"))
+   (logical "c"       "<code>"		"Code"           ("<code>" (r "Code: ") "</code>"))
+   (logical "x"       "<samp>"		"Sample"         ("<samp>" (r "Sample code") "</samp>"))
+   (logical "r"       "<cite>"		"Citation"       ("<cite>" (r "Citation: ") "</cite>"))
+   (logical "k"       "<kbd>"		"Keyboard Input" ("<kbd>" (r "Keyboard: ") "</kbd>"))
+   (logical "v"       "<var>"		"Variable"       ("<var>" (r "Variable: ") "</var>"))
+   (logical "d"       "<dfn>"		"Definition"     ("<dfn>" (r "Definition: ") "</dfn>"))
+   (logical "a"	      "<address>"	"Address"	 ("<address>" r "</address>"))
+   (logical "e"       "<em>"		"Emphasized"     ("<em>" (r "Text: ") "</em>"))
+   (logical "s"       "<strong>"	"Strong"         ("<strong>" (r "Text: ") "</strong>"))
+   (logical "p"       "<pre>"		"Preformatted"   ("<pre>" (r "Text: ") "</pre>"))
 
    ;;physical styles
-   (phys    "-"       "<strike>"	"Strikethru"         ("<strike>" (r "Text: ") "</strike>"))
-   (phys    "u"       "<u>"		"Underline"          ("<u>" (r "Text: ") "</u>"))
-   (phys    "o"       "<i>"		"Italic"             ("<i>" (r "Text: ") "</i>"))
-   (phys    "b"	      "<b>"    		"Bold"               ("<b>" (r "Text: ") "</b>"))
-   (phys    "t"       "<tt>"		"Fixed"              ("<tt>" (r "Text: ") "</tt>"))
+   (phys    "p"       "<sup>"           "Superscript"    ("<sup>" (r "Text: ") "</sup>"))
+   (phys    "u"       "<sub>"           "Subscript"      ("<sub>" (r "Text: ") "</sub>"))
+   (phys    "s"       "<small>"         "Small"          ("<small>" (r "Text: ") "</small>"))
+   (phys    "g"       "<big>"           "Big"            ("<big>" (r "Text: ") "</big>"))
+   (phys    "-"       "<strike>"	"Strikethru"     ("<strike>" (r "Text: ") "</strike>"))
+   (phys    "u"       "<u>"		"Underline"      ("<u>" (r "Text: ") "</u>"))
+   (phys    "o"       "<i>"		"Italic"         ("<i>" (r "Text: ") "</i>"))
+   (phys    "b"	      "<b>"    		"Bold"           ("<b>" (r "Text: ") "</b>"))
+   (phys    "t"       "<tt>"		"Fixed"          ("<tt>" (r "Text: ") "</tt>"))
 
    ;;headers
-   (header  "6"       "<h6>"		"Header 6"       	  ("<h6>" (r "Header: ") "</h6>"))
-   (header  "5"       "<h5>"		"Header 5"       	  ("<h5>" (r "Header: ") "</h5>"))
-   (header  "4"       "<h4>"		"Header 4"       	  ("<h4>" (r "Header: ") "</h4>"))
-   (header  "3"       "<h3>"		"Header 3"       	  ("<h3>" (r "Header: ") "</h3>"))
-   (header  "2"       "<h2>"		"Header 2"       	  ("<h2>" (r "Header: ") "</h2>"))
-   (header  "1"	      "<h1>"     	"Header 1"       	  ("<h1>" (r "Header: ") "</h1>"))
+   (header  "6"       "<h6>"		"Header 6"       ("<h6>" (r "Header: ") "</h6>"))
+   (header  "5"       "<h5>"		"Header 5"       ("<h5>" (r "Header: ") "</h5>"))
+   (header  "4"       "<h4>"		"Header 4"       ("<h4>" (r "Header: ") "</h4>"))
+   (header  "3"       "<h3>"		"Header 3"       ("<h3>" (r "Header: ") "</h3>"))
+   (header  "2"       "<h2>"		"Header 2"       ("<h2>" (r "Header: ") "</h2>"))
+   (header  "1"	      "<h1>"     	"Header 1"       ("<h1>" (r "Header: ") "</h1>"))
 
    ;; forms
-   (form    "o"       "<option>"        "Option"                 (& "<option>" > ))
-   (form    "v"       "<option value"   "Option with Value"      (& "<option value=\"" (r "Value: ") "\">" >))
-   (form    "s"       "<select"		"Selections"	          ("<select name=\"" (p "Name: ") "\">\n<option>" > "\n</select>")"<select")
-   (form    "z"	      "<input"		"Reset Form"    	  ("<input type=\"RESET\" value=\"" (p "Reset button text: ") "\">"))
-   (form    "b"	      "<input"		"Submit Form"   	  ("<input type=\"SUBMIT\" value=\"" (p "Submit button text: ") "\">"))
-   (form    "i"	      "<input"		"Image Field"   	  ("<input type=\"IMAGE\" name=\"" (p "Name: ") "\" src=\"" (p "Image URL: ") "\">"))
-   (form    "h"       "<input"          "Hidden Field"            ("<input type=\"HIDDEN\" name=\"" (p "Name: ") "\" value=\"" (p "Value: ") "\">"))
-   (form    "p"	      "<textarea"	"Text Area"	  ("<textarea name=\"" (p "Name: ") "\" rows=\"" (p "Rows: ") "\" cols=\"" (p "Columns: ") "\">" r "</textarea>"))
-   (form    "c"	      "<input"		"Checkbox"   	          ("<input type=\"CHECKBOX\" name=\"" (p "Name: ") "\">"))
-   (form    "r"	      "<input"		"Radiobutton"   	  ("<input type=\"RADIO\" name=\"" (p "Name: ") "\">"))
-   (form    "t"	      "<input"		"Text Field"	          ("<input type=\"TEXT\" name=\"" (p "Name: ") "\" size=\"" (p "Size: ") "\">"))
-   (form    "f"	      "<form"           "Form"		          ("<form action=\"" (p "Action: ") "\" method=\"" (p "Method: ") "\">\n</form>\n"))
+   (form    "o"       "<option>"        "Option"         (& "<option>" > ))
+   (form    "v"       "<option value"   "Option with Value"  (& "<option value=\"" (r "Value: ") "\">" >))
+   (form    "s"       "<select"		"Selections"	 ("<select name=\"" (p "Name: ") "\">\n<option>" > "\n</select>")"<select")
+   (form    "z"	      "<input"		"Reset Form"     ("<input type=\"RESET\" value=\"" (p "Reset button text: ") "\">"))
+   (form    "b"	      "<input"		"Submit Form"    ("<input type=\"SUBMIT\" value=\"" (p "Submit button text: ") "\">"))
+   (form    "i"	      "<input"		"Image Field"    ("<input type=\"IMAGE\" name=\"" (p "Name: ") "\" src=\"" (p "Image URL: ") "\">"))
+   (form    "h"       "<input"          "Hidden Field"   ("<input type=\"HIDDEN\" name=\"" (p "Name: ") "\" value=\"" (p "Value: ") "\">"))
+   (form    "p"	      "<textarea"	"Text Area"	 ("<textarea name=\"" (p "Name: ") "\" rows=\"" (p "Rows: ") "\" cols=\"" (p "Columns: ") "\">" r "</textarea>"))
+   (form    "c"	      "<input"		"Checkbox"   	 ("<input type=\"CHECKBOX\" name=\"" (p "Name: ") "\">"))
+   (form    "r"	      "<input"		"Radiobutton"    ("<input type=\"RADIO\" name=\"" (p "Name: ") "\">"))
+   (form    "t"	      "<input"		"Text Field"	 ("<input type=\"TEXT\" name=\"" (p "Name: ") "\" size=\"" (p "Size: ") "\">"))
+   (form    "f"	      "<form"           "Form"		 ("<form action=\"" (p "Action: ") "\" method=\"" (p "Method: ") "\">\n</form>\n"))
+
+   ;;tables
+   (table   "d"       "<td>"            "Data cell"      ("<td>"))
+   (table   "h"       "<th>"            "Header"         ("<th>"))
+   (table   "r"       "<tr>"            "Row"            ("<tr>"))
+   (table   "t"       "<table>"         "Table"          ("<table>\n<tr>\n</table>\n"))
 
    ;;lists
-   (list    "t"       "<dt>"            "Definition Item"         (& "<dt>" > (p "Term: ") "\n<dd>" > (r "Definition: ")))
-   (list    "l"       "<li>"            "List Item"               (& "<li>" > (r "Item: ")))
-   (list    "r"	      "<dir>"		"DirectoryList"      	  (& "<dir>" > "\n<li>" > (r "Item: ") "\n</dir>" >))
-   (list    "m"	      "<menu>"		"Menu List"		  (& "<menu>" > "\n<li>" > (r "Item: ") "\n</menu>" >))
-   (list    "o"	      "<ol>"		"Ordered List"   	  (& "<ol>" > "\n<li>" > (r "Item: ") "\n</ol>" >))
-   (list    "d"	      "<dl>"		"Definition List" 	  (& "<dl>" > "\n<dt>" > (p "Term: ") "\n<dd>" > (r "Definition: ") "\n</dl>" >))
-   (list    "u"	      "<ul>"		"Unordered List" 	  (& "<ul>" > "\n<li>" > (r "Item: ") "\n</ul>" >))
+   (list    "t"       "<dt>"            "Definition Item" (& "<dt>" > (p "Term: ") "\n<dd>" > (r "Definition: ")))
+   (list    "l"       "<li>"            "List Item"      (& "<li>" > (r "Item: ")))
+   (list    "r"	      "<dir>"		"DirectoryList"  (& "<dir>" > "\n<li>" > (r "Item: ") "\n</dir>" >))
+   (list    "m"	      "<menu>"		"Menu List"	 (& "<menu>" > "\n<li>" > (r "Item: ") "\n</menu>" >))
+   (list    "o"	      "<ol>"		"Ordered List"   (& "<ol>" > "\n<li>" > (r "Item: ") "\n</ol>" >))
+   (list    "d"	      "<dl>"		"Definition List" (& "<dl>" > "\n<dt>" > (p "Term: ") "\n<dd>" > (r "Definition: ") "\n</dl>" >))
+   (list    "u"	      "<ul>"		"Unordered List" (& "<ul>" > "\n<li>" > (r "Item: ") "\n</ul>" >))
 
    ;;anchors
-   (anchor  "n"	      "<a name="	"Link Target"	  ("<a name=\"" (p "Anchor name: ") "\">" (r "Anchor text: ") "</a>"))
-   (anchor  "h"	      "<a href="        "Hyperlink"          	  ("<a href=\"" (p "URL: ") "\">" (r "Anchor text: ") "</a>"))                
+   (anchor  "n"	      "<a name="	"Link Target"	 ("<a name=\"" (p "Anchor name: ") "\">" (r "Anchor text: ") "</a>"))
+   (anchor  "h"	      "<a href="        "Hyperlink"      ("<a href=\"" (p "URL: ") "\">" (r "Anchor text: ") "</a>"))                
 
    ;;graphics
-   (image   "a"       nil               "Aligned Image"	  ("<img align=\"" (r "Alignment: ") "\" src=\"" (r "Image URL: ") "\">"))
-   (image   "i"       "<img src="	"Image"		  ("<img src=\"" (r "Image URL: ") "\">"))
-   (image   "e"       "<img align="     "Aligned Image With Alt. Text"	  ("<img align=\"" (r "Alignment: ") "\" src=\"" (r "Image URL: ") "\" alt=\"" (r "Text URL: ") "\">"))
-   (image   "t"       "<img alt="	"Image With Alternate Text"	  ("<img alt=\"" (r "Text URL: ") "\" src=\"" (r "Image URL: ") "\">"))
+   (image   "m"       "<map name="      "Image map"      ("<map name=\"" (r "Map name: ") "\">"))
+   (image   "a"       nil               "Aligned Image"	 ("<img align=\"" (r "Alignment: ") "\" src=\"" (r "Image URL: ") "\">"))
+   (image   "i"       "<img src="	"Image"		 ("<img src=\"" (r "Image URL: ") "\">"))
+   (image   "e"       "<img align="     "Aligned Image With Alt. Text"	("<img align=\"" (r "Alignment: ") "\" src=\"" (r "Image URL: ") "\" alt=\"" (r "Text URL: ") "\">"))
+   (image   "t"       "<img alt="	"Image With Alternate Text"	("<img alt=\"" (r "Text URL: ") "\" src=\"" (r "Image URL: ") "\">"))
+
+   ;;specials
+   (special "a"       "<applet code="   "Applet"         ("<applet code=\"" (r "Applet class: ") "\" width=" (r "Applet width: ") " height=" (r "Applet height: ") ">"))
+   (special "b"       "<basefont size=" "Base font size" ("<basefont size=" (r "Font size: ") ">"))
+   (special "c"       "<font color="    "Font color"     ("<font color=\"" (r "Color: ") "\">" (r "Text: ") "</font>"))
+   (special "s"       "<font size="     "Font size"      ("<font size=" (r "Font size: ") ">" (r "Text: ") "</font>"))
 
    ;;text elements
-   (textel  "\C-c="    nil		"Horizontal Line"	  (& "<hr>\n"))
-   (textel  "\C-c\C-m" nil		"Line Break"		  ("<br>\n"))
-   (textel  "\e\C-m"  nil		"Paragraph"	  ("<p>" (progn (sgml-indent-line) nil) "\n"))
+   (textel  "\C-c="   nil		"Horizontal Line" (& "<hr>\n"))
+   (textel  "\C-c\C-m" nil		"Line Break"	 ("<br>\n"))
+   (textel  "\e\C-m"  nil		"Paragraph"	 ("<p>" (progn (sgml-indent-line) nil) "\n"))
 
    ;;head elements
-   (head    "H"       "<head>"          "Head"            ("<head>\n" "</head>\n"))
-   (head    "B"       "<body>"          "Body"            ("<body>\n" "</body>\n"))
-   (head    "i"	      "<isindex>"	"Isindex"         ("<isindex>\n"))
-   (head    "n"	      "<nextid>"	"Nextid"          ("<nextid>\n"))
+   (head    "H"       "<head>"          "Head"           ("<head>\n" "</head>\n"))
+   (head    "B"       "<body>"          "Body"           ("<body>\n" "</body>\n"))
+   (head    "i"	      "<isindex>"	"Isindex"        ("<isindex>\n"))
+   (head    "n"	      "<nextid>"	"Nextid"         ("<nextid>\n"))
    (head    "h"       "<meta http-equiv=" "HTTP Equivalent" ("<meta http-equiv=\"" (p "Equivalent: ") "\" content=\"" (r "Content: ") "\">\n"))
-   (head    "m"       "<meta name="     "Meta Name"       ("<meta name=\"" (p "Name: ") "\" content=\"" (r "Content: ") "\">\n"))
-   (head    "l"	      "<link"		"Link"            ("<link href=\"" p "\">"))
-   (head    "b"       "<base"		"Base"            ("<base href=\"" r "\">"))
-   (head    "t"	      "<title>"		"Title"           ("<title>" (r "Document title: ") "</title>"))
+   (head    "m"       "<meta name="     "Meta Name"      ("<meta name=\"" (p "Name: ") "\" content=\"" (r "Content: ") "\">\n"))
+   (head    "l"	      "<link"		"Link"           ("<link href=\"" p "\">"))
+   (head    "s"       "<script>"        "Script"         ("<script>"))
+   (head    "y"       "<style>"         "Style"          ("<style>"))
+   (head    "b"       "<base"		"Base"           ("<base href=\"" r "\">"))
+   (head    "t"	      "<title>"		"Title"          ("<title>" (r "Document title: ") "</title>"))
    ))
 
 ;;}}}
@@ -483,6 +552,72 @@ appropriate keymap if a key is requested. Format:
 	   (cons '["List Item" html-helper-smart-insert-item t] html-helper-list-menu)))
 
 ;;}}}
+
+;;{{{ menu support
+
+;; menus are built for easymenu. html-helper-add-tag builds
+;; submenus based on tag type, the expert menu code lumps them
+;; together into one list and calls easy-menu-define
+
+(defun html-helper-rebuild-menu nil
+  "Rebuild and install the HTML menu (using `easy-menu-define').
+If `html-helper-use-expert-menu' is nil, then just use a novice menu."
+  (let ((menu (html-helper-expert-menu)))
+    (easy-menu-remove menu)
+    (easy-menu-add menu html-mode-map)))
+
+(defun html-helper-toggle-expert-menu (&optional arg)
+  "Toggle full HTML menus. Optional arg acts like minor-mode args."
+  (interactive "P")
+  (setq html-helper-use-expert-menu
+	(if (null arg) (not html-helper-use-expert-menu)
+	  (> (prefix-numeric-value arg) 0)))
+  (html-helper-rebuild-menu))
+
+;; Expert menus: consed up out of html-helper-installed-types
+(defun html-helper-expert-menu ()
+  "This menu is based on the current value of `html-helper-installed-types'.
+This function can be called again, it redoes the entire menu."
+  ;; Start with the user-provided menu stuff
+  (let ((html-helper-mode-menu html-helper-user-menu))
+    ;; Now cons in the browse-url functions
+    (if (fboundp 'browse-url-of-file)
+	(setq html-helper-mode-menu
+	      (cons '["Load this Buffer in Browser" browse-url-of-file t]
+		    html-helper-mode-menu)))
+    (if (and (boundp 'browse-url-browser-function)
+	     (fboundp browse-url-browser-function))
+	(setq html-helper-mode-menu
+	      (cons (vector "Browse URL at point"
+			    browse-url-browser-function t)
+		    html-helper-mode-menu)))
+
+    ;; cons in the timestamp delimiters
+    (setq html-helper-mode-menu
+	  (cons '["Insert Timestamp Delimiter"
+		  html-helper-insert-timestamp-delimiter-at-point t]
+		html-helper-mode-menu))
+
+    ;; now cons up the main menu out of the submenus
+    (mapcar
+     (function (lambda (type)
+		 (setq html-helper-mode-menu
+		       (cons (html-helper-normalized-menu-for type)
+			     html-helper-mode-menu))))
+     html-helper-installed-types)
+
+    ;; now tack on our name
+    (setq html-helper-mode-menu (cons "Insert" html-helper-mode-menu))
+
+    ;; special mode keys
+    (define-key html-mode-map (kbd "<M-iso-left-tab>") 'tempo-complete-tag)
+    ;;("\M-\C-f" tempo-forward-mark)
+    ;;("\M-\C-b" tempo-backward-mark)
+
+    html-helper-mode-menu))
+
+;;}}}
+
 ;;{{{ patterns for font-lock
 
 ; Old patterns from html-mode.el
@@ -503,7 +638,7 @@ appropriate keymap if a key is requested. Format:
 
 ;; font-lock compatibility with XEmacs/Lucid and older Emacsen (<19.29).
 ;;
-(if (string-match "XEmacs\\|Lucid" (emacs-version))
+(if running-xemacs
     ;; XEmacs/Lucid
     ;; Make needed faces if the user hasn't already done so.
     ;; Respect X resources (`make-face' uses them when they exist).
@@ -647,6 +782,15 @@ appropriate keymap if a key is requested. Format:
 
 ;;}}}
 
+;;{{{ completion finder for tempo
+
+;; The regexp finds everything between the last < or & and point,
+;; which is good enough to match the tags HTML might complete.
+(defvar html-helper-completion-finder "\\(\\(<\\|&\\).*\\)\\="
+  "Passed to tempo-use-tag-list, used to find tags to complete.")
+
+;;}}}
+
 ;;{{{ timestamps
 
 (defun html-helper-update-timestamp ()
@@ -800,9 +944,4 @@ get those characters to appear literally in the output."
   (highlight-headers-follow-url-kfm
    (concat "file:" (buffer-file-name (current-buffer)))))
 
-(define-derived-mode html3-mode html-mode "HTML3"
-  (setq sgml-declaration             (expand-file-name "html3.decl"
-						       sgml-data-directory)
-	sgml-default-doctype-name    "html-3"
-	sgml-shortag                 nil  ))
-
+;;; end of psgml-html.el
