@@ -1,7 +1,7 @@
 ;;; psgml.el --- SGML-editing mode with parsing support
-;; $Id: psgml.el,v 2.20 1996/11/20 18:39:14 lenst Exp $
+;; $Id: psgml.el,v 1.8 2000/03/27 14:10:35 andreasj Exp $
 
-;; Copyright (C) 1993, 1994, 1995, 1996 Lennart Staflin
+;; Copyright (C) 1993-1999 Lennart Staflin
 ;; Copyright (C) 1992 Free Software Foundation, Inc.
 
 ;; Author: Lennart Staflin <lenst@lysator.liu.se>
@@ -52,7 +52,7 @@
 
 ;;; Code:
 
-(defconst psgml-version "1.0.1"
+(defconst psgml-version "1.2.1"
   "Version of psgml package.")
 
 (defconst psgml-maintainer-address "lenst@lysator.liu.se")
@@ -72,8 +72,10 @@
   "Abbrev table in use in sgml-mode.")
 (define-abbrev-table 'sgml-mode-abbrev-table ())
 
-(defvar sgml-running-xemacs
-  (not (not (string-match "Lucid\\|XEmacs" emacs-version))))
+;; XEmacs change
+(if (not (boundp 'running-xemacs))
+    (defvar running-xemacs (string-match "XEmacs\\|Lucid" emacs-version)
+      "Non-nil when the current emacs is XEmacs."))
 
 (defvar sgml-xml-p nil
   "Is this an XML document?")
@@ -100,7 +102,6 @@
   :prefix "sgml-"
   :group 'psgml)
 
-
 (defcustom sgml-insert-missing-element-comment t
   "*If true, and sgml-auto-insert-required-elements also true,
 `sgml-insert-element' will insert a comment if there is an element required
@@ -142,17 +143,16 @@ of a submenu to indicate the range of that menu."
   "*Document type name to use if no document type declaration is present."
   :type '(choice string (const nil))
   :group 'psgml-dtd)
-
 (put 'sgml-default-doctype-name 'sgml-type 'string-or-nil)
 
 (defcustom sgml-markup-faces '((start-tag 	. bold)
-			    (end-tag 	. bold)
-			    (comment 	. italic)
-			    (pi 	. bold)
-			    (sgml 	. bold)
-			    (doctype 	. bold)
-			    (entity 	. bold-italic)
-			    (shortref   . bold))
+			       (end-tag 	. bold)
+			       (comment 	. italic)
+			       (pi 		. bold)
+			       (sgml 		. bold)
+			       (doctype 	. bold)
+			       (entity 		. bold-italic)
+			       (shortref	. bold))
   "*List of markup to face mappings.
 Element are of the form (MARKUP-TYPE . FACE).
 Possible values for MARKUP-TYPE is:
@@ -173,7 +173,8 @@ shortref- short reference"
 (defvar sgml-buggy-subst-char-in-region 
   (or (not (boundp 'emacs-minor-version))
       (not (natnump emacs-minor-version))
-      (< emacs-minor-version 23))
+      (and (eq emacs-major-version 19)
+           (< emacs-minor-version 23)))
   "*If non-nil, work around a bug in subst-char-in-region.
 The bug sets the buffer modified.  If this is set, folding commands
 will be slower.")
@@ -233,14 +234,16 @@ Setting this variable automatically makes it local to the current buffer.")
 
 (defcustom sgml-tag-region-if-active t ;; wing change
   "*If non-nil, the Tags menu will tag a region if the region is 
-considered active by Emacs.  If nil, region must be active and
-transient-mark-mode/zmacs-regions must be on for the region to be tagged."
+considered active by emacs.  If nil, region must be active and
+transient-mark-mode must be on for the region to be tagged."
   :type 'boolean
   :group 'psgml)
 
-(defvar sgml-normalize-trims t
+(defcustom sgml-normalize-trims t
   "*If non-nil, sgml-normalize will trim off white space from end of element
-when adding end tag.")
+when adding end tag."
+  :type 'boolean
+  :group 'psgml)
 
 (defvar sgml-omittag t
   "*Set to non-nil, if you use OMITTAG YES.
@@ -266,6 +269,22 @@ Setting this variable automatically makes it local to the current buffer.")
 (make-variable-buffer-local 'sgml-namecase-general)
 (put 'sgml-namecase-general 'sgml-desc "NAMECASE GENERAL")
 
+
+
+;;[lenst/1998-03-09 19:51:55]
+(defconst sgml-namecase-entity nil)
+
+(defcustom sgml-general-insert-case 'lower
+  "*The case that will be used for general names in inserted markup.
+This can be the symbol `lower' or `upper'. Only effective if
+sgml-namecase-general is true."
+  :type '(choice (const lower) (const upper))
+  :group 'psgml-insert)
+(put 'sgml-general-insert-case 'sgml-type '(lower upper))
+
+(defvar sgml-entity-insert-case nil)
+
+
 (defvar sgml-minimize-attributes nil
   "*Determines minimization of attributes inserted by edit-attributes.
 Actually two things are done
@@ -362,45 +381,52 @@ The tags should be given as strings."
   :type 'boolean
   :group 'psgml)
 
+;;; James addition
+(defvar sgml-menu-name "SGML"
+  "*The name of the menu on which the SGML mode options appear.
+This is intended to be overridden by submodes of sgml-mode.")
+
 ;;; XEmacs change
 (defvar sgml-data-directory (or (locate-data-directory "psgml")
 				(locate-data-directory "sgml"))
   "*Directory for pre-supplied data files (DTD's and such).
 Set this before loading psgml.")
 
-(defcustom sgml-system-path nil
+;; This isn't used now
+(defun sgml-parse-colon-path (cd-path)
+  "Explode a colon-separated list of paths into a string list."
+  (if (null cd-path)
+      nil
+    (let ((cd-sep ":")
+	  cd-list (cd-start 0) cd-colon)
+      (if (boundp 'path-separator)
+	  (setq cd-sep path-separator))
+      (setq cd-path (concat cd-path cd-sep))
+      (while (setq cd-colon (string-match cd-sep cd-path cd-start))
+	(setq cd-list
+	      (nconc cd-list
+		     (list (if (= cd-start cd-colon)
+			       nil
+			     (substitute-in-file-name
+			      (substring cd-path cd-start cd-colon))))))
+	(setq cd-start (+ cd-colon 1)))
+      cd-list)))
+
+(defcustom sgml-system-path (split-path (or (getenv "SGML_SEARCH_PATH") "."))
   ;; wing addition
   "*List of directories used to look for system identifiers.
 The directory listed in `sgml-data-directory' is always searched in
 addition to the directories listed here."
   :type '(repeat directory)
   :group 'psgml)
-(put 'sgml-system-path 'sgml-type 'list)
+(put 'sgml-system-path 'sgml-type 'file-list)
 
-;; This isn't used now
-(defun sgml-parse-colon-path (cd-path)
-  "Explode a colon-separated list of paths into a string list."
-  (let (cd-list (cd-start 0) cd-colon)
-    (setq cd-path (concat cd-path ":"))
-    (while (setq cd-colon (string-match ":" cd-path cd-start))
-      (setq cd-list
-	    (nconc cd-list
-		   (list (if (= cd-start cd-colon)
-			     nil
-			   (substitute-in-file-name
-			    (substring cd-path cd-start cd-colon))))))
-      (setq cd-start (+ cd-colon 1)))
-    cd-list))
-
-(defcustom sgml-public-map (split-path
-			    (or (getenv "SGML_PATH")
-				;; Wing/Krause change
-				(concat "%S"
-					path-separator
-					(directory-file-name
-					 sgml-data-directory)
-					"/%o/%c/%d")))
-  
+(defcustom sgml-public-map (split-path (or (getenv "SGML_PATH")
+					   ;; Wing/Krause change
+					   (concat "%S" path-separator
+						   (directory-file-name
+						    sgml-data-directory)
+						   "%o/%c/%d")))
   "*Mapping from public identifiers to file names.
 This is a list of possible file names.  To find the file for a public
 identifier the elements of the list are used one at the time from the
@@ -425,13 +451,12 @@ This variable is automatically local to the buffer."
   :type '(repeat file)
   :group 'psgml-dtd)
 (make-variable-buffer-local 'sgml-local-catalogs)
-(put 'sgml-local-catalogs 'sgml-type 'list)
+(put 'sgml-local-catalogs 'sgml-type 'file-list)
 
 (defcustom sgml-catalog-files (split-path
 			       (or (getenv "SGML_CATALOG_FILES")
 				   ;; Wing/Krause addition
-				   (concat "CATALOG"
-					   path-separator
+				   (concat "CATALOG" path-separator
 					   (expand-file-name
 					    "CATALOG"
 					    sgml-data-directory))))
@@ -440,7 +465,7 @@ The files are in the format defined in the SGML Open Draft Technical
 Resolution on Entity Management."
   :type '(repeat file)
   :group 'psgml-dtd)
-(put 'sgml-catalog-files 'sgml-type 'list)
+(put 'sgml-catalog-files 'sgml-type 'file-list)
 
 ;;; Wing addition
 (defcustom sgml-ecat-files (list
@@ -450,21 +475,24 @@ Resolution on Entity Management."
   "*List of catalog files for PSGML."
   :type '(repeat file)
   :group 'psgml-dtd)
-(put 'sgml-ecat-files 'sgml-type 'list)
+(put 'sgml-ecat-files 'sgml-type 'file-list)
 
 (defcustom sgml-local-ecat-files nil
   "*List of local catalog files for PSGML.
 Automatically becomes buffer local if set."
   :type '(repeat file)
   :group 'psgml-dtd)
-(make-variable-buffer-local 'sgml-local-ecat-files)
-(put 'sgml-local-ecat-files 'sgml-type 'list)
 
-(defvar sgml-public-transliterations '((?  . ?_) (?/ . ?%) (?: . ?-))
+(make-variable-buffer-local 'sgml-local-ecat-files)
+(put 'sgml-local-ecat-files 'sgml-type 'file-list)
+
+(defcustom sgml-public-transliterations '((? . ?_) (?/ . ?%) (?: . ?-))
   "*Transliteration for characters that should be avoided in file names.
 This is a list of dotted pairs (FROM . TO); where FROM is the the
 character to be translated to TO.  This is used when parts of a public
-identifier are used to construct a file name.")
+identifier are used to construct a file name."
+  :type '(repeat (cons character character))
+  :group 'psgml-dtd)
 
 (defvar sgml-default-dtd-file nil
   "*This is the default file name for saved DTD.
@@ -484,9 +512,9 @@ use `setq-default' to set it to a value that is shared among buffers.")
 (put 'sgml-exposed-tags 'sgml-type 'list)
 
 
-(defvar sgml-custom-markup nil
+(defcustom sgml-custom-markup nil
   "*Menu entries to be added to the Markup menu.
-The value should be a list of lists of two strings.  The first is a
+The value should be a list of lists of two strings.  The first
 string is the menu line and the second string is the text inserted
 when the menu item is chosen.  The second string can contain a \\r
 where the cursor should be left.  Also if a selection is made
@@ -498,7 +526,11 @@ Example:
 
   ((\"Version1\" \"<![%Version1[\\r]]>\")
    (\"New page\"  \"<?NewPage>\"))
-")
+"
+  :type '(repeat (list :inline t
+		       (string :tag "Menu Line")
+		       (string :tag "Inserted String")))
+  :group 'psgml)
 
 (defcustom sgml-custom-dtd nil
   "Menu entries to be added to the DTD menu.
@@ -530,7 +562,6 @@ Example:
 				     (symbol :tag "Variable")
 				     (sexp :tag "Value")))))
   :group 'psgml-dtd)
-			       
 
 
 ;;; Faces used in edit attribute buffer:
@@ -544,8 +575,9 @@ Example:
 ;;; The -s option suppresses output.
 
 (defcustom sgml-validate-command (concat "nsgmls -s -m "
-					 sgml-data-directory
-					 "/CATALOG %s %s")
+					 (expand-file-name "CATALOG"
+							   sgml-data-directory)
+					 " %s %s")
   "*The shell command to validate an SGML document.
 
 This is a `format' control string that by default should contain two
@@ -568,14 +600,17 @@ string will be replaced according to the list below, if the string contains
 "
   :type 'string
   :group 'psgml)
+(make-variable-buffer-local 'sgml-validate-command)
 
-(defvar sgml-validate-files nil
+(defcustom sgml-validate-files nil
   "If non-nil, a function of no arguments that returns a list of file names.
 These file names will serve as the arguments to the `sgml-validate-command'
-format control string instead of the defaults.")
+format control string instead of the defaults."
+  :type 'hook
+  :group 'psgml)
 
 (defvar sgml-validate-error-regexps
-  '((":\\(.+\\):\\([0-9]+\\):\\([0-9]+\\):[EX]: " 1 2 3)
+  '((".*:\\(.+\\):\\([0-9]+\\):\\([0-9]+\\):[EX]: " 1 2 3)
     ("\\(error\\|warning\\) at \\([^,]+\\), line \\([0-9]+\\)" 2 3)
     ("\n[a-zA-Z]?:?[^0-9 \n\t:]+:[ \t]*\\([^ \n\t:]+\\):\
 \\([0-9]+\\):\\(\\([0-9]+\\)[: \t]\\)?" 1 2 4))
@@ -584,13 +619,15 @@ See `compilation-error-regexp-alist'.")
 
 (defcustom sgml-declaration nil
   "*If non-nil, this is the name of the SGML declaration file."
-  :type 'hook
+  :type 'file
   :group 'psgml-dtd)
-(put 'sgml-declaration 'sgml-type 'string)
+(put 'sgml-declaration 'sgml-type 'file-or-nil)
 
-(defvar sgml-xml-declaration nil
-  "*If non-nil, this is the name of the SGML declaration for XML files.")
-(put 'sgml-xml-declaration 'sgml-type 'string)
+(defcustom sgml-xml-declaration nil
+  "*If non-nil, this is the name of the SGML declaration for XML files."
+  :type 'file
+  :group 'psgml-dtd)
+(put 'sgml-xml-declaration 'sgml-type 'file-or-nil)
 
 (defcustom sgml-mode-hook nil
   "A hook or list of hooks to be run when entering sgml-mode"
@@ -602,6 +639,7 @@ See `compilation-error-regexp-alist'.")
     sgml-omittag
     sgml-shorttag
     sgml-namecase-general
+    sgml-general-insert-case
     sgml-minimize-attributes
     sgml-always-quote-attributes
     sgml-indent-step
@@ -644,6 +682,7 @@ See `compilation-error-regexp-alist'.")
     sgml-public-map
     sgml-catalog-files
     sgml-ecat-files
+    sgml-general-insert-case
     )
   "User options that can be saved or set from menu."
   )
@@ -698,7 +737,9 @@ See `compilation-error-regexp-alist'.")
 	     (insert
 	      "<!-- Keep this comment at the end of the file\n"
 	      "Local variables:\n"
-	      "mode: sgml\n"
+	      (if sgml-xml-p
+		  "mode: xml\n"
+		"mode: sgml\n")
 	      "End:\n"
 	      "-->\n")
 	     (forward-line -3)))
@@ -751,7 +792,7 @@ as that may change."
 	 (symbol-value hook)
 	 (let ((value (symbol-value hook)))
 	   (if (and (listp value) (not (eq (car value) 'lambda)))
-	       (mapcar '(lambda (foo) (apply foo args))
+	       (mapcar (function (lambda (foo) (apply foo args)))
 		       value)
 	     (apply value args))))))
 
@@ -762,9 +803,9 @@ as that may change."
 
 (defun sgml-markup (entry text)
   (cons entry
-	(` (lambda ()
-	     (interactive)
-	     (sgml-insert-markup (, text))))))
+	(` (function (lambda ()
+		       (interactive)
+		       (sgml-insert-markup (, text)))))))
 
 (defun sgml-insert-markup (text)
   (let ((end (sgml-mouse-region))
@@ -786,7 +827,7 @@ as that may change."
 (defun sgml-mouse-region ()
   (let (start end)
     (cond
-     (sgml-running-xemacs
+     (running-xemacs
       (cond
        ((null (mark-marker)) nil)
        (t (setq start (region-beginning)
@@ -827,7 +868,8 @@ as that may change."
        (reporter-submit-bug-report
 	psgml-maintainer-address
 	(concat "psgml.el " psgml-version)
-	(list 
+	(list
+	 'major-mode
 	 'sgml-always-quote-attributes
 	 'sgml-auto-activate-dtd
 	 'sgml-auto-insert-required-elements
@@ -853,6 +895,7 @@ as that may change."
 	 'sgml-shorttag
 	 'sgml-namecase-general
 	 'sgml-tag-region-if-active
+         'sgml-use-text-properties
 	 ))))
 
 
@@ -864,64 +907,64 @@ as that may change."
 
 ;;; Key commands
 
-(define-key sgml-mode-map [(tab)]    'sgml-indent-or-tab)
-;;;(define-key sgml-mode-map [(?<)] 'sgml-insert-tag)
-(define-key sgml-mode-map [(?>)]     'sgml-close-angle)
-(define-key sgml-mode-map [(?/)]     'sgml-slash)
-(define-key sgml-mode-map [(control ?c) (?#)]    'sgml-make-character-reference)
-(define-key sgml-mode-map [(control ?c) (?-)]    'sgml-untag-element)
-(define-key sgml-mode-map [(control ?c) (?+)]    'sgml-insert-attribute)
-(define-key sgml-mode-map [(control ?c) (?/)]    'sgml-insert-end-tag)
-(define-key sgml-mode-map [(control ?c) (?<)]    'sgml-insert-tag)
-(define-key sgml-mode-map [(control ?c) (?=)]    'sgml-change-element-name)
-(define-key sgml-mode-map [(control ?c) (control ?a)] 'sgml-edit-attributes)
-(define-key sgml-mode-map [(control ?c) (control ?c)] 'sgml-show-context)
-(define-key sgml-mode-map [(control ?c) (control ?d)] 'sgml-next-data-field)
-(define-key sgml-mode-map [(control ?c) (control ?e)] 'sgml-insert-element)
-(define-key sgml-mode-map [(control ?c) (control ?k)] 'sgml-kill-markup)
-(define-key sgml-mode-map [(control ?c) (control ?l)] 'sgml-show-or-clear-log)
-(define-key sgml-mode-map [(control ?c) (control ?n)] 'sgml-up-element)
-(define-key sgml-mode-map [(control ?c) (control ?o)] 'sgml-next-trouble-spot)
-(define-key sgml-mode-map [(control ?c) (control ?p)] 'sgml-parse-prolog)
-(define-key sgml-mode-map [(control ?c) (control ?q)] 'sgml-fill-element)
-(define-key sgml-mode-map [(control ?c) (control ?r)] 'sgml-tag-region)
-(define-key sgml-mode-map [(control ?c) (control ?s)] 'sgml-unfold-line)
-(define-key sgml-mode-map [(control ?c) (control ?t)] 'sgml-list-valid-tags)
-(define-key sgml-mode-map [(control ?c) (control ?v)] 'sgml-validate)
-(define-key sgml-mode-map [(control ?c) (control ?w)] 'sgml-what-element)
-(define-key sgml-mode-map [(control ?c) (control ?z)] 'sgml-trim-and-leave-element)
-(define-key sgml-mode-map [(control ?c) (control ?f) (control ?e)] 'sgml-fold-element)
-(define-key sgml-mode-map [(control ?c) (control ?f) (control ?r)] 'sgml-fold-region)
-(define-key sgml-mode-map [(control ?c) (control ?f) (control ?s)] 'sgml-fold-subelement)
-(define-key sgml-mode-map [(control ?c) (control ?f) (control ?x)] 'sgml-expand-element)
-(define-key sgml-mode-map [(meta control ?O)] 'sgml-split-element)
-(define-key sgml-mode-map [(control ?c) (control ?u) (control ?e)] 'sgml-unfold-element)
-(define-key sgml-mode-map [(control ?c) (control ?u) (control ?a)] 'sgml-unfold-all)
-(define-key sgml-mode-map [(control ?c) (control ?u) (control ?l)] 'sgml-unfold-line)
-(define-key sgml-mode-map [(control ?c) (control ?u) (control ?d)] 'sgml-custom-dtd)
-(define-key sgml-mode-map [(control ?c) (control ?u) (control ?m)] 'sgml-custom-markup)
+(define-key sgml-mode-map "\t"    'sgml-indent-or-tab)
+;(define-key sgml-mode-map "<" 	  'sgml-insert-tag)
+(define-key sgml-mode-map ">"     'sgml-close-angle)
+(define-key sgml-mode-map "/"     'sgml-slash)
+(define-key sgml-mode-map "\C-c#"    'sgml-make-character-reference)
+(define-key sgml-mode-map "\C-c-"    'sgml-untag-element)
+(define-key sgml-mode-map "\C-c+"    'sgml-insert-attribute)
+(define-key sgml-mode-map "\C-c/"    'sgml-insert-end-tag)
+(define-key sgml-mode-map "\C-c<"    'sgml-insert-tag)
+(define-key sgml-mode-map "\C-c="    'sgml-change-element-name)
+(define-key sgml-mode-map "\C-c\C-a" 'sgml-edit-attributes)
+(define-key sgml-mode-map "\C-c\C-c" 'sgml-show-context)
+(define-key sgml-mode-map "\C-c\C-d" 'sgml-next-data-field)
+(define-key sgml-mode-map "\C-c\C-e" 'sgml-insert-element)
+(define-key sgml-mode-map "\C-c\C-f\C-e" 'sgml-fold-element)
+(define-key sgml-mode-map "\C-c\C-f\C-r" 'sgml-fold-region)
+(define-key sgml-mode-map "\C-c\C-f\C-s" 'sgml-fold-subelement)
+(define-key sgml-mode-map "\C-c\C-f\C-x" 'sgml-expand-element)
+(define-key sgml-mode-map "\C-c\C-i" 'sgml-add-element-to-element)
+(define-key sgml-mode-map "\C-c\C-k" 'sgml-kill-markup)
+(define-key sgml-mode-map "\C-c\C-l" 'sgml-show-or-clear-log)
+(define-key sgml-mode-map "\C-c\r"   'sgml-split-element)
+(define-key sgml-mode-map "\C-c\C-n" 'sgml-up-element)
+(define-key sgml-mode-map "\C-c\C-o" 'sgml-next-trouble-spot)
+(define-key sgml-mode-map "\C-c\C-p" 'sgml-parse-prolog)
+(define-key sgml-mode-map "\C-c\C-q" 'sgml-fill-element)
+(define-key sgml-mode-map "\C-c\C-r" 'sgml-tag-region)
+(define-key sgml-mode-map "\C-c\C-s" 'sgml-unfold-line)
+(define-key sgml-mode-map "\C-c\C-t" 'sgml-list-valid-tags)
+(define-key sgml-mode-map "\C-c\C-u\C-a" 'sgml-unfold-all)
+(define-key sgml-mode-map "\C-c\C-u\C-d" 'sgml-custom-dtd)
+(define-key sgml-mode-map "\C-c\C-u\C-e" 'sgml-unfold-element)
+(define-key sgml-mode-map "\C-c\C-u\C-l" 'sgml-unfold-line)
+(define-key sgml-mode-map "\C-c\C-u\C-m" 'sgml-custom-markup)
+(define-key sgml-mode-map "\C-c\C-v" 'sgml-validate)
+(define-key sgml-mode-map "\C-c\C-w" 'sgml-what-element)
+(define-key sgml-mode-map "\C-c\C-z" 'sgml-trim-and-leave-element)
 
-(define-key sgml-mode-map [(meta control ?a)]   'sgml-beginning-of-element)
-(define-key sgml-mode-map [(meta control ?e)]   'sgml-end-of-element)
-(define-key sgml-mode-map [(meta control ?f)]   'sgml-forward-element)
-(define-key sgml-mode-map [(meta control ?b)]   'sgml-backward-element)
-(define-key sgml-mode-map [(meta control ?d)]   'sgml-down-element)
-(define-key sgml-mode-map [(meta control ?u)]   'sgml-backward-up-element)
-(define-key sgml-mode-map [(meta control ?k)]   'sgml-kill-element)
-(define-key sgml-mode-map [(meta control ?@)]   'sgml-mark-element)
-(define-key sgml-mode-map [(meta control ?h)]   'sgml-mark-current-element)
-(define-key sgml-mode-map [(meta control ?t)]   'sgml-transpose-element)
-(define-key sgml-mode-map [(meta tab)]		'sgml-complete)
+(define-key sgml-mode-map "\e\C-a"   'sgml-beginning-of-element)
+(define-key sgml-mode-map "\e\C-e"   'sgml-end-of-element)
+(define-key sgml-mode-map "\e\C-f"   'sgml-forward-element)
+(define-key sgml-mode-map "\e\C-b"   'sgml-backward-element)
+(define-key sgml-mode-map "\e\C-d"   'sgml-down-element)
+(define-key sgml-mode-map "\e\C-u"   'sgml-backward-up-element)
+(define-key sgml-mode-map "\e\C-k"   'sgml-kill-element)
+(define-key sgml-mode-map "\e\C-@"   'sgml-mark-element)
+;;(define-key sgml-mode-map [?\M-\C-\ ] 'sgml-mark-element)
+(define-key sgml-mode-map "\e\C-h"   'sgml-mark-current-element)
+(define-key sgml-mode-map "\e\C-t"   'sgml-transpose-element)
+(define-key sgml-mode-map "\M-\t"    'sgml-complete)
 
 ;;;; Menu bar
 
 (easy-menu-define
  sgml-dtd-menu sgml-mode-map "DTD menu"
- '("DTD"))
-
-(defconst sgml-dtd-root-menu
-  '("DTD"
+ '("DTD"
     ["Parse DTD"  sgml-parse-prolog t]
+    ("Insert DTD")
     ("Info"
      ["General DTD info"	sgml-general-dtd-info           t]
      ["Describe element type"	sgml-describe-element-type	t]
@@ -935,7 +978,7 @@ as that may change."
     "--"
     ["Load Parsed DTD"  sgml-load-dtd t]
     ["Save Parsed DTD"  sgml-save-dtd t]
-    ))
+   ))
 
 (easy-menu-define
  sgml-view-menu sgml-mode-map "View menu"
@@ -950,23 +993,22 @@ as that may change."
    ["Hide Tags"		sgml-hide-tags		t]
    ["Hide Attributes"	sgml-hide-attributes	t]
    ["Show All Tags"	sgml-show-tags		t]
-   )
- )
+   ))
 
 
-(defconst sgml-markup-root-menu
-  '("Markup"
-    ["Insert Element"	sgml-element-menu	t]
-    ["Insert Start-Tag" sgml-start-tag-menu	t]
-    ["Insert End-Tag"	sgml-end-tag-menu	t]
-    ["Tag Region"	sgml-tag-region-menu	t]
-    ["Insert Attribute" sgml-attrib-menu	t]
-    ["Insert Entity"	sgml-entities-menu	t]
-    ))
-
 (easy-menu-define
  sgml-markup-menu sgml-mode-map "Markup menu"
- sgml-markup-root-menu)
+ '("Markup"
+   ["Insert Element"	sgml-element-menu	t]
+   ["Insert Start-Tag" sgml-start-tag-menu	t]
+   ["Insert End-Tag"	sgml-end-tag-menu	t]
+   ["End Current Element"	sgml-insert-end-tag t]
+   ["Tag Region"	sgml-tag-region-menu	t]
+   ["Insert Attribute"  sgml-attrib-menu	t]
+   ["Insert Entity"	sgml-entities-menu	t]
+   ["Add Element to Element"	sgml-add-element-menu	t]
+   ("Custom markup"   "---")
+   ))
 
 (easy-menu-define
  sgml-move-menu sgml-mode-map "Menu of move commands"
@@ -1007,43 +1049,115 @@ as that may change."
  sgml-main-menu sgml-mode-map "Main menu"
  '("SGML"
    ["Reset Buffer"	normal-mode t]
-   ["End Element"	sgml-insert-end-tag t]
    ["Show Context"	sgml-show-context t]
    ["What Element"	sgml-what-element t]
    ["List Valid Tags"	sgml-list-valid-tags t]
    ["Show/Hide Warning Log"  sgml-show-or-clear-log t]
    ["Validate"		sgml-validate t]
-   ["File Options >"	sgml-file-options-menu t]
-   ["User Options >"	sgml-user-options-menu t]
-   ["Save File Options"  sgml-save-options t]
+   ("File Options"   "---")
+   ("User Options"   "---")
    ["Submit Bug Report"  sgml-submit-bug-report t]
-   )
- )
+   ))
 
 
+(defun sgml-options-menu-items (vars)
+  (if running-xemacs
+      (sgml-make-options-menu vars)
+    (mapcar (function
+	     (lambda (var)
+	       (let ((desc (format "%s [%s]"
+				   (sgml-variable-description var)
+				   (sgml-option-value-indicator var)))
+		     (type (sgml-variable-type var)))
+		 (cond ((consp type)
+			(cons desc
+			      (mapcar (function
+				       (lambda (c)
+					 (vector
+					  (if (consp c)
+					      (car c)
+					    (format "%s" c))
+					  `(setq ,var
+						 ',(if (consp c) (cdr c) c))
+					  t))
+				       type))))
+		       (t
+			(vector desc `(sgml-do-set-option ',var) t)))))
+	     vars))))
+
+(defun sgml-option-value-indicator (var)
+  (let ((type (sgml-variable-type var))
+	(val (symbol-value var)))
+    (cond
+     ((eq type 'toggle)
+      (if val "Yes" "No"))
+     ((eq type 'string)
+      (if (stringp val)
+	  (substring val 0 (min (length val) 4))
+	"-"))
+     ((and (atom type) val)
+      "...")
+     ((consp type)
+      (or (car (rassq val type))
+	  val))
+     (t
+      "-"))))
+
+(defvar sgml-last-options-menu-values ())
+
+(defun sgml-any-option-changed (oldvalues vars)
+  (not (loop for val in oldvalues
+             for var in vars
+             always (eq val (symbol-value var)))))
+
+(defun sgml-update-options-menu (menuname option-vars &optional save-func)
+  (let ((last-values (assoc menuname sgml-last-options-menu-values)))
+    (when (or (null last-values)
+              (sgml-any-option-changed (cdr last-values)
+                                       option-vars))
+      (easy-menu-change (list sgml-menu-name) menuname
+                      (nconc (sgml-options-menu-items option-vars)
+                             (if save-func
+                                 (list "---"
+                                       (vector (format "Save %s" menuname)
+                                               save-func t)))))
+      (unless last-values
+        (setq last-values (cons menuname nil))
+        (push last-values sgml-last-options-menu-values))
+      (setf (cdr last-values) (mapcar (function symbol-value) option-vars)))))
+
+
+(defun sgml-update-all-options-menus ()
+  (sgml-update-options-menu "File Options" sgml-file-options
+			    'sgml-save-options)
+  (sgml-update-options-menu "User Options" sgml-user-options)
+  nil)
+
+(defun sgml-compute-insert-dtd-items ()
+  (loop for e in sgml-custom-dtd collect
+        (vector (first e)
+                (` (sgml-doctype-insert (, (cadr e)) '(, (cddr e))))
+                t)))
+
+(defun sgml-compute-custom-markup-items ()
+  (loop for e in sgml-custom-markup collect
+        (vector (first e)
+                (` (sgml-insert-markup  (, (cadr e))))
+                t)))
+
 (defun sgml-build-custom-menus ()
   "Build custom parts of Markup and DTD menus."
   (let ((button3 (lookup-key (current-local-map) [button3])))
-    (easy-menu-define
-     sgml-markup-menu sgml-mode-map "Markup menu"
-     (append sgml-markup-root-menu
-	     (list "----")
-	     (loop for e in sgml-custom-markup collect
-		   (vector (first e)
-			   (` (sgml-insert-markup  (, (cadr e))))
-			   t))))
-    (easy-menu-define
-     sgml-dtd-menu sgml-mode-map "DTD menu"
-     (append sgml-dtd-root-menu
-	     (list "----")
-	     (loop for e in sgml-custom-dtd collect
-		   (vector (first e)
-			   (` (sgml-doctype-insert (, (cadr e))
-						   '(, (cddr e))))
-			   t))))    
     (unless (or (null button3)
 		(numberp button3))
-      (local-set-key [button3] button3))))
+      (local-set-key [button3] button3))
+    (when sgml-custom-dtd
+      (easy-menu-change '("DTD") "Insert DTD"
+			(sgml-compute-insert-dtd-items)))
+    (when sgml-custom-markup
+      (easy-menu-change '("Markup") "Custom markup"
+			(sgml-compute-custom-markup-items))))
+  nil)
 
 
 ;;;; Post command hook 
@@ -1067,15 +1181,19 @@ actually only the state that persists between commands.")
 
 
 (defun sgml-command-post ()
-  (when (eq major-mode 'sgml-mode)
+  (when (or (memq major-mode '(sgml-mode xml-mode))
+	    (fboundp 'make-local-hook))
+    ;; Explanation if make-local-hook is defined then this is called
+    ;; from a local hook and need not check major-mode.
     (when (and (null sgml-buffer-parse-state)
 	       sgml-auto-activate-dtd
 	       (null sgml-auto-activate-dtd-tried)
 	       (not (zerop (buffer-size)))
 	       (looking-at ".*<"))
       (setq sgml-auto-activate-dtd-tried t)
-      (sgml-need-dtd)
-      (sgml-fontify-buffer 0))
+      (ignore-errors
+        (sgml-need-dtd)
+        (sgml-fontify-buffer 0)))
     (when sgml-buffer-parse-state
       (sgml-update-display))))
 
@@ -1170,7 +1288,7 @@ All bindings:
 
   (set (make-local-variable 'paragraph-separate)
 	"^[ \t\n]*$\\|\
-^[ \t]*</?\\([A-Za-z]\\([-.A-Za-z0-9= \t\n]\\|\
+^[ \t]*</?\\([_A-Za-z]\\([-:._A-Za-z0-9= \t\n]\\|\
 \"[^\"]*\"\\|'[^']*'\\)*\\)?>$")
   (set (make-local-variable 'paragraph-start)
        paragraph-separate)
@@ -1193,25 +1311,39 @@ All bindings:
   (make-local-variable 'mode-line-format)
   ;; Modify mode-line-format with susbt (sugested by wing)
   (setq mode-line-format
-	(subst '("" mode-name sgml-active-dtd-indicator) 'mode-name
-	       mode-line-format))
+ 	(subst '("" mode-name sgml-active-dtd-indicator) 'mode-name
+ 	       mode-line-format))
   (make-local-variable 'sgml-default-dtd-file)
   (when (setq sgml-default-dtd-file (sgml-default-dtd-file))
     (unless (file-exists-p sgml-default-dtd-file)
       (setq sgml-default-dtd-file nil)))
-  (add-hook 'post-command-hook 'sgml-command-post 'append)
+  (cond ((fboundp 'make-local-hook)
+	 ;; emacs>= 19.29
+	 (make-local-hook 'post-command-hook)
+	 (add-hook 'post-command-hook 'sgml-command-post 'append 'local)
+	 (make-local-hook 'activate-menubar-hook)
+         (add-hook 'activate-menubar-hook 'sgml-update-all-options-menus
+		   nil 'local))
+	(t
+	 ;; emacs< 19.29
+	 (add-hook 'post-command-hook 'sgml-command-post 'append)
+         (add-hook 'menu-bar-update-hook 'sgml-update-all-options-menus)
+         ))
   (run-hooks 'text-mode-hook 'sgml-mode-hook)
-  (sgml-build-custom-menus)
   (easy-menu-add sgml-main-menu)
   (easy-menu-add sgml-modify-menu)
   (easy-menu-add sgml-move-menu)
   (easy-menu-add sgml-markup-menu)
   (easy-menu-add sgml-view-menu)
-  (easy-menu-add sgml-dtd-menu))
+  (easy-menu-add sgml-dtd-menu)
+  (sgml-build-custom-menus))
+
+
 
 ;;;###autoload
 (define-derived-mode xml-mode sgml-mode "XML"
   (setq sgml-xml-p t)
+  ;; XML-friendly settings
   (setq sgml-omittag nil)
   (setq sgml-shorttag nil)
   (setq sgml-namecase-general nil)
@@ -1219,6 +1351,7 @@ All bindings:
   (setq sgml-always-quote-attributes t)
   (setq sgml-validate-command "nsgmls -wxml -s %s %s")
   (unless sgml-declaration
+    (make-local-variable 'sgml-declaration)
     (setq sgml-declaration sgml-xml-declaration)))
 
 (defun sgml-default-dtd-file ()
@@ -1241,7 +1374,7 @@ All bindings:
     0))
 
 (defconst sgml-start-tag-regex
-  "<[A-Za-z]\\([-.A-Za-z0-9= \n\t]\\|\"[^\"]*\"\\|'[^']*'\\)*"
+  "<[_A-Za-z]\\([-:.A-Za-z0-9= \n\t]\\|\"[^\"]*\"\\|'[^']*'\\)*"
   "Regular expression that matches a non-empty start tag.
 Any terminating > or / is not matched.")
 
@@ -1309,7 +1442,7 @@ Any terminating > or / is not matched.")
 		    ;; nor an unclosed start-tag.
 		    (looking-at (concat sgml-start-tag-regex "[/<]"))
 		    ;; Nor an unclosed end-tag.
-		    (looking-at "</[A-Za-z][-.A-Za-z0-9]*[ \t]*<"))
+		    (looking-at "</[A-Za-z][-:.A-Za-z0-9]*[ \t]*<"))
 		   (setq blinkpos nil)))
 	    (if blinkpos
 		()
@@ -1542,6 +1675,9 @@ declaration names.
 If it is something else complete with ispell-complete-word." t)
 (autoload 'sgml-file-options-menu "psgml-edit" nil t)
 (autoload 'sgml-user-options-menu "psgml-edit" nil t)
+(autoload 'sgml-add-element-to-element "psgml-edit" "Add an element of type GI to the current element.
+The element will be added at the last legal position if FIRST is `nil',
+otherwise it will be added at the first legal position." t)
 (autoload 'sgml-save-dtd "psgml-dtd" "Save the parsed dtd on FILE." t)
 (autoload 'sgml-list-elements "psgml-info" "List the elements and their attributes in the current DTD." t)
 (autoload 'sgml-list-attributes "psgml-info" "List the attributes and in which elements they occur." t)
@@ -1556,11 +1692,12 @@ If it is something else complete with ispell-complete-word." t)
 
 
 ;;;; Last provisions
+
 (provide 'psgml)
 (provide 'sgml-mode)
 
 (cond
- (sgml-running-xemacs
+ (running-xemacs
   (require 'psgml-xemacs))
  (t
   (require 'psgml-other)))
